// SPDX-License-Identifier: GPL-2.0
/*
 * bcachefs module version information.
 *
 * Copyright 2010, 2011 Kent Overstreet <kent.overstreet@gmail.com>
 * Copyright 2012 Google, Inc.
 */

#include <linux/module.h>
#include "version.h"

MODULE_VERSION(bcachefs_version);
